import logging
import os
import textwrap

import yaml
from deepmerge import Merger
from jinja2 import Environment, Undefined
from yaml.parser import ParserError
from yaml.scanner import ScannerError

logger = logging.getLogger(__name__)


class SilentUndefined(Undefined):
    """Do not error out if a value is not found"""

    def _fail_with_undefined_error(self, *args, **kwargs):
        return ""


def jinja_dict_to_yaml(*obj, **options) -> str:
    """
    Convert dict to yaml in human readable format.

    This writes the yaml in full format, for better parsing of jinja includes

    Args:
        obj: the object to present as yaml
        options: optional parameters to pass on to yaml.safe_dump function

    Return:
        A human readable yaml presentation of the original input

    """
    if options.get('indent'):
        indent = options['indent']
        del options['indent']
        return textwrap.indent(yaml.safe_dump(*obj, indent=2, allow_unicode=True, default_flow_style=False, **options),
                               ' ' * indent)

    return yaml.safe_dump(*obj, indent=4, allow_unicode=True, default_flow_style=False, **options)


class ParseJinja:
    """Parser for all jinja related files, and allows you to merge the output with initial parameters"""

    def __init__(self):
        self.path = os.path.dirname(os.path.realpath(__file__))
        self.jinja2env = Environment(undefined=SilentUndefined)
        self.jinja2env.filters['to_yaml'] = jinja_dict_to_yaml

    @staticmethod
    def merge(parameters_left: dict, parameters_right: dict) -> dict:
        """Combines 2 dicts with a merger

        Args:
            parameters_left (dict): low prio dict
            parameters_right (dict): high prio dict

        Returns:
            dict: merged dict
        """
        my_merger = Merger(
            [  # merger strategy
                (list, ["append"]),
                (dict, ["merge"]),
            ],
            ["override"],  # fallback strategies,
            ["override"],  # conflict strategy
        )
        return my_merger.merge(parameters_left, parameters_right)

    @staticmethod
    def get_yaml_filenames(path: str) -> list[str]:
        """get all yaml files in specified path

        Args:
            path (str): path to a directory

        Returns:
            list[str]: list of yaml files found
        """
        return [os.path.join(path, filename) for filename in os.listdir(f"{path}") if filename.endswith(".yaml")]

    def parse_directory(self, path: str, parameters: dict = None) -> dict:
        """parse all yaml files in the directory and return their merged value as dict
            results get merged with parameters before return

        Args:
            path (str): path to a directory
            parameters (dict, optional): parameters to merge the dict with. Defaults to {}.

        Returns:
            dict: merged dict of all parsed files and provided parameters
        """
        if parameters is None:
            parameters = {}
        for filename in self.get_yaml_filenames(path=path):
            parameters = self.parse_file(filename=filename, parameters=parameters)
        return parameters

    def parse_file(self, filename: str, parameters=None) -> dict:
        """parse a specific file and merge its result with parameters

        Args:
            filename (str): _description_
            parameters (dict, optional): parameters to merge the dict with. Defaults to {}.

        Raises:
            FileNotFoundError: file was not found

        Returns:
            dict: merged dict of all parsed file and provided parameters
        """
        if parameters is None:
            parameters = {}
        print(f"parsing file: {filename}")
        with open(filename, encoding="utf-8") as file:
            j2 = file.read()
            output = self.jinja2env.from_string(j2).render(env=os.environ, **parameters)
            logger.debug(f"output: {output}")
            try:
                y = yaml.safe_load(output)
            except ParserError as e:
                raise ParserError(f"in content:\n{output}\nerror: {e}")
            except ScannerError as e:
                raise ParserError(f"in content:\n{output}\nerror: {e}")
            if not y:
                return parameters
            return self.merge(parameters, y)



import pytest

from src.terrajinja.cli.parse_args import ParseArgs, dotdict, parse_variables, KeyNotInPath


class ParseArgsNew(ParseArgs):
    def parse_args(self):
        """return all current parsable arguments"""
        return self.parser.parse_args()


class TestParseArgs:
    def test_dotdict_valid(self):
        path = "deep.deeper.right_string"
        dictionary = {
            'right_string': 'my_right',
            'shared_array': ["2", "3"],
            'deep': {
                'deeper': {
                    'right_string': 'my_right',
                    'shared_array': ["2", "3"]
                }
            }
        }
        result = dotdict(path, dictionary)
        assert result == 'my_right'

    def test_dotdict_invalid(self):
        path = "deep.non_existing.right_string"
        dictionary = {
            'right_string': 'my_right',
            'shared_array': ["2", "3"],
            'deep': {
                'deeper': {
                    'right_string': 'my_right',
                    'shared_array': ["2", "3"]
                }
            }
        }

        with pytest.raises(KeyNotInPath) as context:
            dotdict(path, dictionary)

        assert f"unable to parse dotted string" in str(context.value)

    def test_parse_variables_valid(self):
        path = "key1.$key2.key3.$key4"
        dictionary = {
            'key1': 'my_key1',
            'key2': 'my_key2',
            'key3': 'my_key3',
            'key4': 'my_key4',
        }
        result = parse_variables(path, dictionary)
        # only $ based keys should have been replaces with vars in dict
        assert result == 'key1.my_key2.key3.my_key4'

    def test_parse_variables_invalid(self):
        path = "key1.$key2.key3.$unaffected"
        dictionary = {
            'key1': 'my_key1',
            'key2': 'my_key2',
            'key3': 'my_key3',
            'key4': 'my_key4',
        }
        result = parse_variables(path, dictionary)
        # only $ based keys should have been replaces with vars in dict
        # unmatched keys are ignored
        assert result == 'key1.my_key2.key3.$unaffected'


if __name__ == "__main__":
    pytest.main()



#!/usr/bin/env python

import logging

import yaml

from . import ParseConfig
from terrajinja.deploy import TerraformDeployment


def main():
    logger = logging

    parser = ParseConfig()
    parameters = parser.get_config()

    logger.basicConfig(
        format="%(asctime)s %(message)s",
        level=logging.getLevelName(parameters.get("loglevel").upper()),
    )

    logger.debug(yaml.safe_dump(parameters))

    terraform_config = parameters.get("terraform")
    if terraform_config:
        app = TerraformDeployment(name=parameters["deployment"])
        app.compile(
            terraform_config.get("providers"), terraform_config.get("resources")
        )
        app.action(parameters["action"])



import os

import pytest

from src.terrajinja.cli.parse_jinja import ParseJinja


class TestParseJinja:
    dir = os.path.dirname(os.path.abspath(__file__))

    def test_merge(self):
        left = {
            'left_string': 'my_left',
            'shared_array': ["1", "2"],
            'deep': {
                'deeper': {
                    'left_string': 'my_left',
                    'shared_array': ["1", "2"]
                }
            }
        }
        right = {
            'right_string': 'my_right',
            'shared_array': ["2", "3"],
            'deep': {
                'deeper': {
                    'right_string': 'my_right',
                    'shared_array': ["2", "3"]
                }
            }
        }
        parser = ParseJinja()
        merged = parser.merge(left, right)

        assert merged['left_string'] == 'my_left'
        assert merged['right_string'] == 'my_right'
        assert merged['deep']['deeper']['left_string'] == 'my_left'
        assert merged['deep']['deeper']['right_string'] == 'my_right'
        assert merged['shared_array'] == ["1", "2", "2", "3"]

    def test_get_yaml_filenames(self):
        parser = ParseJinja()
        files = parser.get_yaml_filenames(os.path.join(self.dir, 'config', 'parameters'))
        assert os.path.join(self.dir, 'config', 'parameters', 'test.yaml') in files

    def test_parse_file_parameters(self):
        parser = ParseJinja()
        y = parser.parse_file(os.path.join(self.dir, 'config', 'parameters', 'test.yaml'))
        assert len(y['array_of_choices']) == 3
        assert y['test_var'] == "my_var"
        assert y['path']['to']['custom_choice']['choices'][0] == "choice_A"

    def test_parse_file_templates(self):
        parser = ParseJinja()
        y = parser.parse_file(os.path.join(self.dir, 'config', 'templates', 'default_template_v0.0.1.yaml'),
                              {'test_var': 'my_test_var'})
        assert y['template']['my_string'] == "my_test_var"

    def test_parse_directory(self):
        parser = ParseJinja()
        y = parser.parse_directory(os.path.join(self.dir, 'config', 'parameters'))
        assert len(y['array_of_choices']) == 6
        assert y['test_var'] == "my_var"
        assert y['path']['to']['custom_choice']['choices'][0] == "choice_A"


if __name__ == "__main__":
    pytest.main()



import os

import pytest

from src.terrajinja.cli.parser import ParseConfig


class TestParser:
    dir = os.path.dirname(os.path.abspath(__file__))

    def test_parse_templates_order_1(self):
        parser = ParseConfig(custom_args=["deploy"])
        parameters = {
            'templates': {
                "default_template": "0.0.1",
                "additional_template": "0.0.1"
            }

        }
        dictionary = parser.parse_templates(os.path.join(self.dir, 'config', 'templates'), parameters)
        assert dictionary['original'] == 'additional'

    def test_parse_templates_order_2(self):
        parser = ParseConfig(custom_args=["deploy"])
        parameters = {
            'templates': {
                "additional_template": "0.0.1",
                "default_template": "0.0.1"
            }

        }
        dictionary = parser.parse_templates(os.path.join(self.dir, 'config', 'templates'), parameters)
        assert dictionary['original'] == 'original'


if __name__ == "__main__":
    pytest.main()



import logging
import os
import shutil

from .parse_args import ParseArgs
from .parse_jinja import ParseJinja

logger = logging.getLogger(__name__)


class ParseConfig:
    def __init__(self, custom_args=None):
        self.jinja = ParseJinja()
        self.parse_args = ParseArgs(self.jinja, custom_args=custom_args)

    def get_config(self) -> dict:
        """parses all config files

        Returns:
            dict: merge of all configs

        """
        # get initial arguments and read parameters
        args, _ = self.parse_args.parse_known_args()

        if args.command == "template":
            self.use_template(args)
            exit(0)

        if args.command == "init":
            self.init_dir(args)
            exit(0)

        parameter_path = os.path.join(args.config_directory, 'parameters')
        deployments_path = os.path.join(args.config_directory, 'deployments')
        templates_path = os.path.join(args.config_directory, 'templates')

        # read all parameter files
        parameters = self.jinja.parse_directory(path=parameter_path)
        # if parameters is empty, no parameters where provided in yaml, odd but sure, your new
        if not parameters:
            print(f'warn: no parameters were read from the parameter path {parameter_path}, this is unexpected')

        # read the deployment arg
        deployment = self.parse_args.get_deployment(path=deployments_path, parameters=parameters)

        # parse deployment input requirements and enhance parameters with the result
        parameters = self.parse_args.parse_deployment_input(deployment=deployment, parameters=parameters)
        if not parameters:
            print('error: no parameters were read from the provided deployment file, please provide a yaml file as input')
            exit(1)

        # re-read the deployment file so that its jinja parsed too for the templates etc
        parameters = self.jinja.parse_file(filename=os.path.join(deployments_path, f"{parameters['deployment']}.yaml"),
                                           parameters=parameters)

        # finally parse the templates
        parameters = self.parse_templates(templates_path, parameters)
        return parameters

    def parse_templates(self, path: str, parameters: dict) -> dict:
        """parse all template files using the parameters provided

        Args:
            path: directory to the parameter files
            parameters: dict to use as input for the parameter files

        Returns:
            dict: formatted templates merged in a single dict

        """
        templates = parameters.get("templates")
        if templates:
            for template_filename, version in templates.items():
                parameters = self.jinja.parse_file(
                    filename=os.path.join(path, f"{template_filename}_v{version}.yaml",
                                          ),
                    parameters=parameters,
                )

        return parameters

    @staticmethod
    def init_dir(args: any):
        action = False
        for sub in ["deployments", "templates", "parameters"]:
            _path = os.path.join(args.config_directory, sub)
            if not os.path.exists(_path):
                print('creating directory', _path)
                os.mkdir(_path)
                action = True

        if not action:
            print('init already done, have a look at the template sub command to add a template (tjcli template -l)')
            exit(1)

        print(
            'initial directories created, have a look at the template sub command to add a template (tjcli template -l)')

    @staticmethod
    def use_template(args: any):
        template_dir = os.path.join(os.path.dirname(__file__), '..', 'templates')

        if args.list:
            print('available templates:')
            for root, dirs, filenames in os.walk(template_dir):
                if len(dirs) == 1:
                    description = open(os.path.join(template_dir, dirs[0], 'description.txt')).read()
                    print(f'  {dirs[0]}: {description}')

        if args.add:
            print('requesting to add template {args.template}...')
            template_named_dir = os.path.join(template_dir, args.add)
            if not os.path.exists(template_named_dir):
                print(f'error: non-existing template: {args.add}')
                exit(1)

            # check if destination path exists
            for sub in ["deployments", "templates", "parameters"]:
                destination_path = os.path.join(args.config_directory, sub)
                if not os.path.exists(destination_path):
                    print(f'destination path: {destination_path} does not exist, did you run tjcli init?')
                    exit(1)

            # copy all files from template
            for sub in ["deployments", "templates", "parameters"]:
                template_final_dir = os.path.join(template_named_dir, sub)
                destination_path = os.path.join(args.config_directory, sub)

                # check if template has sub dir
                if os.path.exists(template_final_dir):
                    for filename in os.listdir(template_final_dir):
                        destination_file = os.path.join(destination_path, filename)
                        print(f'dest: {destination_file}')
                        if os.path.exists(destination_file) and not args.force:
                            print(f'file {destination_file} already exists, aborting. use --force to overwrite')
                            exit(1)

                    for filename in os.listdir(template_final_dir):
                        print(f'adding: {sub} / {filename}')
                        shutil.copyfile(os.path.join(template_final_dir, filename),
                                        os.path.join(destination_path, filename))



from setuptools import setup, find_packages, find_namespace_packages
import os

requirements = [line.strip()
          for line in open('requirements.txt').readlines()
          if line.strip() and not line.startswith('#')]

readme = open('README.md').read()
history = open('HISTORY.md').read()
version = '0.1.0'
name= 'terrajinja-cli'
package_path= name.replace('-', '.')

# only the cli has an entry point
entry_points=None
if name=='terrajinja-cli':
    entry_points={
      'console_scripts': [
        'tjcli = terrajinja.cli:main',
      ]
    }

setup(
    name=name,
    version=version,
    description='''Terrajinja extension for automation using cdktf''',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    url='https://gitlab/terrajinja',
    author='Terrajinja Team',
    author_email='int-terrajinja@schubergphilis.com',
    license='MIT',
    include_package_data=True,
    install_requires=requirements,
    packages=find_namespace_packages(where='src/', include=[package_path]),
    package_dir={'': 'src'},
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Natural Language :: English',
      'Programming Language :: Python :: 3.11',
    ],
    entry_points=entry_points,
)



import argparse
import logging
import os
import sys

logger = logging.getLogger(__name__)


class KeyNotInPath(Exception):
    """key does not exist in path"""


def dotdict(dotpath: str, dictionary: dict) -> any:
    """Get the path from a dict based on a dotted path notation

    Args:
        dotpath (str): path of the variable to fetch
        dictionary (dict): dict to fetch the variable from

    Raises:
        ValueError: _description_

    Returns:
        any: result of the item in dict
    """
    path = dotpath.split(".")
    if not dictionary.get(path[0]):
        raise KeyNotInPath(f"unable to parse dotted string '{dotpath}' does not exist in the parameters")
    if len(path) > 1:
        return dotdict(".".join(path[1:]), dictionary[path[0]])
    return dictionary[path[0]]


def parse_variables(input_string: str, parameters: dict) -> str:
    """replace variables in a string based on a dictionary
        replaces strings that start with a '$' with the value of the key with this name

    Args:
        input_string (str): string that contains variables
        parameters (dict): dictionary of variables and values to apply

    Returns:
        str: formatted string
    """
    temp = []
    for word in input_string.split("."):
        if word[0] == "$":
            word = parameters.get(word[1:], word)
        temp.append(word)
    res = ".".join(temp)
    return res


class ParseArgs:
    """Parse commandline arguments, and generate choices based on existing files"""

    def __init__(self, jinja: any, parameters: dict = None, custom_args=None):
        if parameters is None:
            parameters = {}
        self.jinja = jinja
        self.parameters = parameters
        self.parser = argparse.ArgumentParser()
        self.sub_parser = self.parser.add_subparsers(dest="command")
        self.deploy_parser = self.sub_parser.add_parser("deploy", help='start a deployment')
        self.template_parser = self.sub_parser.add_parser("template", help='list ot get templates to start with')
        self.init_parser = self.sub_parser.add_parser("init", help='initialize configuration structure')

        # default parameters which are always available
        self.parser.add_argument(
            "-C",
            "--config-directory",
            # default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "config"),
            default=os.path.join(os.getcwd()),
            help="path to the config directory (default: current working directory)",
        )
        # self.deploy_parser.add_argument(
        #     "-A",
        #     "--action",
        #     required=False,
        #     choices=["plan", "apply", "destroy"],
        #     help="Action to perform on the deployment",
        # )
        self.deploy_parser.add_argument(
            "-l",
            "--loglevel",
            choices=["debug", "info", "warning", "error", "critical"],
            help="Set the logging level (default: %(default)s)",
            default="INFO",
        )
        # template
        self.template_parser.add_argument(
            "-l",
            "--list",
            action='store_true',
            help="list available templates",
        )
        self.template_parser.add_argument(
            "-a",
            "--add",
            help="add a template",
        )
        self.template_parser.add_argument(
            "-f",
            "--force",
            action='store_true',
            help="overwrite existing files when adding a template",
        )

        args, _ = self.parser.parse_known_args(custom_args)
        if len(sys.argv) == 1:
            self.parser.print_help()
            exit(1)

        if args.command == "init":
            self.parser.parse_args(custom_args)

        if args.command == "template":
            self.parser.parse_args(custom_args)

        if args.command == "template" and len(sys.argv) == 2:
            self.template_parser.print_help()
            exit(1)

    @staticmethod
    def get_file_base_names_from_path(path: str) -> list[str]:
        """get available deployments"""
        return [x.split(".")[0] for x in os.listdir(path)]

    def get_deployment(self, path: str, parameters: dict) -> dict:
        if len(self.get_file_base_names_from_path(path)) == 0:
            print(f"No deployments available in {path}. Create one first or start by using a template (tjcli template)")
            exit(1)

        """parse argument of deployment, and get additional arguments from the deployment yaml"""
        self.deploy_parser.add_argument(
            "-d",
            "--deployment",
            required=True,
            choices=self.get_file_base_names_from_path(path),
            help="Name of the deployment",
        )
        args, _unknown = self.parser.parse_known_args()
        return self.jinja.parse_file(os.path.join(path, f"{args.deployment}.yaml"))

    def parse_deployment_input(self, deployment, parameters) -> dict:
        required_input = deployment.get("required_input")
        if not required_input:
            return parameters

        for arg, config in required_input.items():
            choices = config.get("choices")
            if choices and not isinstance(choices, list):
                parameter = dotdict(parse_variables(choices, parameters), parameters)
                if isinstance(parameter, dict):
                    config["choices"] = [x for x in parameters.get(choices).keys()]
                if isinstance(parameter, str):
                    config["choices"] = [parameters]

            default = config.get("default")
            if default:
                if "." in default:
                    config["default"] = dotdict(parse_variables(default, parameters), parameters)

            self.parser.add_argument(f"-{arg[0]}", f"--{arg}", **config)

            # after each arg, parse it, so it can be used as input in the next if needed, merge it in parameters
            args, _unknown = self.parser.parse_known_args()
            parameters = self.jinja.merge(vars(args), parameters)
        return parameters

    def parse_args(self) -> argparse:
        """return all current parsable arguments"""
        return self.parser.parse_args()

    def parse_known_args(self) -> argparse:
        """return all current parsable arguments"""
        return self.parser.parse_known_args()
